/*
 * readanysf~  external for pd. 
 * 
 * Copyright (C) 2003 August Black
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ReadFlac.cpp
 */
#ifdef READ_FLAC

#include <m_pd.h>
#include "ReadFlac.h"

extern "C" {
#include "FLAC/metadata.h"
};



ReadFlac::ReadFlac() {
  needs_seek = false;
  seek_sample = 0;
  samples_in_reservoir =0;
  abort_flag = false;
  decoder = NULL;
}

ReadFlac::~ReadFlac() {
  cleanup();
}

bool ReadFlac::Initialize(const char *filename) {

  if (fp != 0) {
    fclose(fp); fprintf(stderr, "already opened, now closing file");
  }
  fp = fopen( filename,  "r"); 

  //@@@ to be really "clean" we should go through the reader instead of directly to the file...
  if(!FLAC__metadata_get_streaminfo(filename, &streaminfo)) {
    //post("what the fuck");
    return 1;
  }
  
  //length_msec = lengthInMsec();
  /*post("FLAC:<%ihz:%ibps:%dch>", 
       streaminfo.data.stream_info.sample_rate, 
       streaminfo.data.stream_info.bits_per_sample, 
       streaminfo.data.stream_info.channels); //@@@ fix later
  */

  samplerate = (double)streaminfo.data.stream_info.sample_rate;
  num_channels = streaminfo.data.stream_info.channels;
  lengthinseconds = streaminfo.data.stream_info.total_samples/samplerate;

  decoder = FLAC__seekable_stream_decoder_new();
  if(decoder == 0)
    return false;
  FLAC__seekable_stream_decoder_set_md5_checking(decoder, false);
  FLAC__seekable_stream_decoder_set_read_callback(decoder, readCallback_);
  FLAC__seekable_stream_decoder_set_seek_callback(decoder, seekCallback_);
  FLAC__seekable_stream_decoder_set_tell_callback(decoder, tellCallback_);
  FLAC__seekable_stream_decoder_set_length_callback(decoder, lengthCallback_);
  FLAC__seekable_stream_decoder_set_eof_callback(decoder, eofCallback_);
  FLAC__seekable_stream_decoder_set_write_callback(decoder, writeCallback_);
  FLAC__seekable_stream_decoder_set_metadata_callback(decoder, metadataCallback_);
  FLAC__seekable_stream_decoder_set_error_callback(decoder, errorCallback_);
  FLAC__seekable_stream_decoder_set_client_data(decoder, this);
    
  if(FLAC__seekable_stream_decoder_init(decoder) != FLAC__SEEKABLE_STREAM_DECODER_OK) {
    cleanup();
    return false;
  }
  if(!FLAC__seekable_stream_decoder_process_until_end_of_metadata(decoder)) {
    cleanup();
    return false;
  }
  
  return true;
}



int ReadFlac::Decode(float *buffer, int size) {
  
  if(decoder == NULL)
    return 0;
  
  if(needs_seek) {
    FLAC__seekable_stream_decoder_seek_absolute(decoder, seek_sample);
    //post("seeking %d samples", seek_sample);
    needs_seek = false;
  }
  
  
  while(samples_in_reservoir < 576) {
    if(FLAC__seekable_stream_decoder_get_state(decoder) == FLAC__SEEKABLE_STREAM_DECODER_END_OF_STREAM) {
      //post("FLAC: end of file");
      return 0;
    }
    else if(!FLAC__seekable_stream_decoder_process_single(decoder)) {
      //post("FLAC: no process single");
      return 0;
    }
  }

  int n = samples_in_reservoir; // > 576 ? samples_in_reservoir: 576;
  const unsigned channels = streaminfo.data.stream_info.channels;

  if(samples_in_reservoir == 0) {
    //post("FLAC: reservoir is empty");
    return 0;
  }  else {
    
    //const unsigned bits_per_sample = streaminfo.data.stream_info.bits_per_sample;
    //const unsigned bytes_per_sample = (bits_per_sample+7)/8;
    //const unsigned sample_rate = streaminfo.data.stream_info.sample_rate;
    unsigned i;
    //16 > WHDR2 + 2 ? 16 : WHDR2 + 2
    
    //unsigned  delta;
    

    for(i = 0; i < n*channels; i++)
      buffer[i] = (float) ( reservoir[i]/ 32768.0 );
    

    samples_in_reservoir = 0;
    
    //const int bytes = n * channels * bytes_per_sample;
  }
  
  if(eof)
    return 0;
  
  return n*channels; //1;
}

bool ReadFlac::Rewind() {
  needs_seek = true;
  seek_sample = 0;
  return true;
}

bool ReadFlac::PCM_seek(long bytes) {
  needs_seek = true;
  seek_sample = bytes;
  return true;
}

bool ReadFlac::TIME_seek(double seconds) {
  needs_seek = true;
  //lengthInMsec();
  seek_sample = (FLAC__uint64)(seconds * streaminfo.data.stream_info.sample_rate);
  return true;
}


void ReadFlac::cleanup()
{
  if(decoder) {
    FLAC__seekable_stream_decoder_finish(decoder);
    FLAC__seekable_stream_decoder_delete(decoder);
    decoder = NULL;
  }
}

FLAC__SeekableStreamDecoderReadStatus ReadFlac::readCallback_(const FLAC__SeekableStreamDecoder *decoder, 
							     FLAC__byte buffer[], 
							     unsigned *bytes, 
							     void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  *bytes = fread( (char*)buffer, sizeof(unsigned char), *bytes, instance->getFp());
  //*bytes = instance->reader->read((char*)buffer, *bytes);
  return FLAC__SEEKABLE_STREAM_DECODER_READ_STATUS_OK;
}

FLAC__SeekableStreamDecoderSeekStatus ReadFlac::seekCallback_(const FLAC__SeekableStreamDecoder *decoder, 
							      FLAC__uint64 absolute_byte_offset, 
							      void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  if ( fseek( instance->getFp(), (long)absolute_byte_offset, SEEK_SET) == 0 ) {
    //post("seeked %ld bytes", absolute_byte_offset);
    return FLAC__SEEKABLE_STREAM_DECODER_SEEK_STATUS_OK;    
  }else {
    //post("COULD NOT seek %ld bytes", absolute_byte_offset);
    return FLAC__SEEKABLE_STREAM_DECODER_SEEK_STATUS_ERROR;

  }
}


FLAC__SeekableStreamDecoderTellStatus ReadFlac::tellCallback_(const FLAC__SeekableStreamDecoder *decoder, 
							     FLAC__uint64 *absolute_byte_offset, 
							     void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  long pos = ftell( instance->getFp() );
  if(pos < 0)
    return FLAC__SEEKABLE_STREAM_DECODER_TELL_STATUS_ERROR;
  else {
    *absolute_byte_offset = pos;
    return FLAC__SEEKABLE_STREAM_DECODER_TELL_STATUS_OK;
  }
}

FLAC__SeekableStreamDecoderLengthStatus ReadFlac::lengthCallback_(const FLAC__SeekableStreamDecoder *decoder,
								 FLAC__uint64 *stream_length, 
								 void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  long length = ftell( instance->getFp() ); 
  if(length < 0)
    return FLAC__SEEKABLE_STREAM_DECODER_LENGTH_STATUS_ERROR;
  else {
    *stream_length = length;
    return FLAC__SEEKABLE_STREAM_DECODER_LENGTH_STATUS_OK;
  }
}

FLAC__bool ReadFlac::eofCallback_(const FLAC__SeekableStreamDecoder *decoder, 
				  void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  return feof( instance->getFp() );
}

FLAC__StreamDecoderWriteStatus ReadFlac::writeCallback_(const FLAC__SeekableStreamDecoder *decoder, 
							const FLAC__Frame *frame, 
							const FLAC__int32 * const buffer[], 
							void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  //const unsigned bps = instance->streaminfo.data.stream_info.bits_per_sample;
  const unsigned channels = instance->streaminfo.data.stream_info.channels;
  const unsigned wide_samples = frame->header.blocksize;
  unsigned wide_sample, sample, channel;
  
  (void)decoder;
  
  if(instance->abort_flag)
    return FLAC__STREAM_DECODER_WRITE_STATUS_ABORT;
  
  for(sample = instance->samples_in_reservoir*channels, wide_sample = 0; wide_sample < wide_samples; wide_sample++)
    for(channel = 0; channel < channels; channel++, sample++)
      instance->reservoir[sample] = (FLAC__int16)buffer[channel][wide_sample];
  
  instance->samples_in_reservoir += wide_samples;
  
  return FLAC__STREAM_DECODER_WRITE_STATUS_CONTINUE;
}

void ReadFlac::metadataCallback_(const FLAC__SeekableStreamDecoder *decoder, 
				const FLAC__StreamMetadata *metadata, 
				void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  (void)decoder;
  if(metadata->type == FLAC__METADATA_TYPE_STREAMINFO) {
    instance->streaminfo = *metadata;
    
    if(instance->streaminfo.data.stream_info.bits_per_sample != 16) {
      instance->abort_flag = true;
      return;
    }
  }
}

void ReadFlac::errorCallback_(const FLAC__SeekableStreamDecoder *decoder, 
			     FLAC__StreamDecoderErrorStatus status, 
			     void *client_data) {
  ReadFlac *instance = (ReadFlac*)client_data;
  (void)decoder;
  if(status != FLAC__STREAM_DECODER_ERROR_STATUS_LOST_SYNC)
    instance->abort_flag = true;
}

#endif
