/*
 * readanysf~  external for pd. 
 * 
 * Copyright (C) 2003 August Black
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ReadMadUrl.cpp  || much code (C) by Olaf Matthes
 * http://www.akustische-kunst.org/puredata/shout/shout.html
 */
#ifdef READ_MAD_URL

#include <m_pd.h>
#include "ReadMadUrl.h"

#include <ctype.h>  // isdigit

#define mad_f_tofloat(x)  ((float) ((x) / (float) (1L << MAD_F_FRACBITS)))
#define     sys_closesocket close
#define     STRDUP strdup
#define     STRBUF_SIZE             1024



int receive(int fd, unsigned char *rcvbuffer, int size) {
  int   ret = -1;
  ret = recv(fd, rcvbuffer, size, 0);
  if(ret < 0)
    {
      post("receive error" );
    }
  return ret;
}

void  *fill_infifo(void *zz ) {
  int ret, wret, tmpfd, count=0;
  unsigned char tmp[SOCKET_READSIZE];
  ReadMadUrl *rmu = (ReadMadUrl *)zz;

  while (1){ 
    pthread_mutex_lock( rmu->get_mutex() );
    if ( rmu->get_quit() ) {
      pthread_mutex_unlock( rmu->get_mutex() );
      break;
    }

    while ( rmu->get_fifo()->FreeSpace() > SOCKET_READSIZE+576) {  
      if ( rmu->get_quit() ) {
	break;
      }
      tmpfd = rmu->get_fd();
      rmu->set_cachesize( ); //(float)rmu->get_fifo()->madfifo->UsedSpace();  
      pthread_mutex_unlock( rmu->get_mutex() );
      ret = receive( tmpfd , tmp, SOCKET_READSIZE );
      //post("receiving data");
      pthread_mutex_lock( rmu->get_mutex() );

      if (ret > 0) {
	wret = rmu->get_fifo()->Write( (void *)tmp, ret*sizeof(unsigned char)); 
      } else {
	bzero((void *)tmp, sizeof(tmp)); //write silence to the fifo
	wret = rmu->get_fifo()->Write( (void *)tmp, ret*sizeof(unsigned char)); 
	if (count++ > 5) {
	  count = 0;
	  sys_closesocket(tmpfd);
	  post( "closed the connection, reconnecting to %s, %s, %d", 
		rmu->get_hostname(), 
		rmu->get_mountpoint(), 
		rmu->get_port() );
	  if (( tmpfd = rmu->socket_connect( rmu->get_hostname(), 
					     rmu->get_mountpoint(), 
					     rmu->get_port() ) ) < 0) {	  
	    post("couldn't reconnect,...trying again");
	  }
	  rmu->set_fd(tmpfd);
	} else {
	  post("couldn't get data, waiting %d seconds.", 6-count);
	  break; //or sleep
	}
      }
    }
    if ( rmu->get_quit() ) {
      //post("got quit,...now quitting.");
      pthread_mutex_unlock( rmu->get_mutex() );
      break;
    }
    pthread_mutex_unlock( rmu->get_mutex() );
    //post("fifo full in thread.. waiting");
    sleep(1);//pthread_cond_wait( rmu->get_condition() , rmu->get_mutex() );
  }


  post("exiting thread");
  pthread_cond_signal(rmu->get_condition());
  return NULL;
}



ReadMadUrl::ReadMadUrl() {
  mad_fd = -1;
  port = 0;
  bytes_avail=0;
  madfifo = new Fifo( STREAM_FIFOSIZE );
  quit = false;

  pthread_mutex_init(&x_mutex, 0);
  pthread_cond_init(&x_condition, 0);

}

ReadMadUrl::~ReadMadUrl() {

  pthread_mutex_lock(&x_mutex);
  quit = true;
  pthread_mutex_unlock(&x_mutex);
  post("signalling...");
  pthread_cond_signal(&x_condition);

  pthread_cond_wait(&x_condition  , &x_mutex );
  //post("after waiting...");

  if (mad_fd >= 0)
    socket_disconnect(mad_fd);
  //  if (pthread_join(x_childthread, &threadrtn))
  //error("readsf_free: join failed");

  post("... done.");
  
  if ( mad_init) {
    mad_synth_finish (&synth);
    mad_frame_finish (&frame);
    mad_stream_finish(&stream);
    mad_init = 0;
  }

  pthread_cond_destroy(&x_condition);
  pthread_mutex_destroy(&x_mutex);
  
}				

bool ReadMadUrl::fill_buffer( ) {
  size_t bytes_read;

  memmove(mad_map, mad_map + STREAM_BUFFERSIZE - bytes_avail, bytes_avail);
  bytes_read = STREAM_BUFFERSIZE - bytes_avail;
  
  pthread_mutex_lock(&x_mutex);
  madfifo->Read( (void *)(mad_map + bytes_avail),  bytes_read);
  pthread_mutex_unlock(&x_mutex);
 
  //post("RMU cachesize %f", cachesize);
  map_offset += (STREAM_BUFFERSIZE - bytes_avail);
  bytes_avail += bytes_read;

  return 1;
}


int ReadMadUrl::Decode(float *buffer, int size) {
  
  struct mad_pcm *pcm;
  mad_fixed_t const *left_ch;
  mad_fixed_t const *right_ch;
  int nsamples;
  int nchannels;

  /*if ( mad_fd < 0) {
    post( "reconnecting to %s, %s, %d", hostname, mountpoint, port);
    if (( mad_fd = socket_connect( hostname, mountpoint, port)) < 0) {
    return 0;  //return 0;
    }
    }
  */
  if ( bytes_avail < 3072) {
    fill_buffer( );
    mad_stream_buffer(&stream, mad_map, bytes_avail);
  } else {
    //post("bytes_avail = %d", bytes_avail);
  }



  if (mad_frame_decode(&frame, &stream) == -1) {
    if (!MAD_RECOVERABLE( stream.error)) {
      mad_frame_mute(&frame);
      return 0;
    } 
    mad_frame_mute(&frame);
    //else {
    //if ( feof( mad_fd) ) {
    //post("MAD: end of file");
    //return 0;
    //}	
    //post("MAD error: (not fatal)"); 
  }
  

  mad_synth_frame (&synth, &frame);
  
  {
    pcm = &synth.pcm;
    nsamples = pcm->length;
    nchannels = pcm->channels;
    left_ch = pcm->samples[0];
    right_ch = pcm->samples[1];
    int x = 0;
    while (nsamples--) {
      buffer[x++] = mad_f_tofloat(*(left_ch++));
      if (nchannels == 2) 
	buffer[x++] = mad_f_tofloat(*(right_ch++));
    }
  }

  bytes_avail = stream.bufend - stream.next_frame;
  //if (pcm->length != 1152) 
  //post(" %d", pcm->length);
  return pcm->length*num_channels;
}




ssize_t ReadMadUrl::find_initial_frame(uint8_t *buf, int size) {
  uint8_t *data = buf;
  int ext_header = 0;
  int pos = 0;
  ssize_t header_size = 0;
  while (pos < (size - 10)) {
    if (pos == 0 && data[pos] == 0x0d && data[pos+1] == 0x0a)
      pos += 2;
    if (data[pos] == 0xff && (data[pos+1] == 0xfb
			      || data[pos+1] == 0xfa
			      || data[pos+1] == 0xf3 
			      || data[pos+1] == 0xe2
			      || data[pos+1] == 0xe3)) {
      //post("found header at %d", pos);
      return pos;
    }	
    if (pos == 0 && data[pos] == 0x0d && data[pos+1] == 0x0a) {
      return -1; // Let MAD figure this out
    }	
    if (pos == 0 && (data[pos] == 'I' && data[pos+1] == 'D' && data[pos+2] == '3')) {
      header_size = (data[pos + 6] << 21) + 
	(data[pos + 7] << 14) +
	(data[pos + 8] << 7) +
	data[pos + 9]; /* syncsafe integer */
      if (data[pos + 5] & 0x10) {
	ext_header = 1;
	header_size += 10; /* 10 byte extended header */
      }

      header_size += 10;
      
      if (header_size > STREAM_BUFFERSIZE) {
	//post("Header larger than 32K (%d)", header_size);
	return header_size;
      }	
      return header_size;
    } else if (pos == 0 && data[pos] == 'T' && data[pos+1] == 'A' && 
	       data[pos+2] == 'G') {
      return 128;	/* TAG is fixed 128 bytes, we assume! */
    }  else {
      pos++;
    }				
  }
  post("MAD debug: potential problem file or unhandled info block");
  post("next 4 bytes =  %x %x %x %x (index = %d, size = %d)",
	 data[header_size], data[header_size+1],
	 data[header_size+2], data[header_size+3],
	 header_size, size);
  return -1;
}

bool ReadMadUrl::Rewind() {
  /* if (mad_fd >= 0) 
     { socket_disconnect(mad_fd); mad_fd =-1; }
  */
  //post("ReadMadUrl::rewind ??");
  return false;
}

bool ReadMadUrl::Initialize( const char *fullurl ) {

  set_url(fullurl);
  
  if (( mad_fd = socket_connect( hostname, mountpoint, port)) < 0) {
    post( "Initialized() failed");
    return false;  //return 0;
  }

  //post("ReadMadUrl::got socket connection...");

  seekable = 0;  // NOT seekable 
  mad_synth_init  (&synth);
  mad_stream_init (&stream);
  mad_frame_init  (&frame);
  mad_init = 1;
  
  //  post("ReadMadUrl::filling  in fifo...");
  //if ( ! fill_infifo() )
  //return 0;
  //post("ReadMadUrl::filling in buffer...");
  //if ( !fill_buffer( ) )
  //{ socket_disconnect(mad_fd); mad_fd =-1; return 0; }


  unsigned char tmp[SOCKET_READSIZE];
  size_t bytes_read;
  //post("ReadMadUrl::getting first frame...");
  int ret = receive( mad_fd, tmp, SOCKET_READSIZE );
  //post("ReadMadUrl::received some data...");

  if (ret > 0) {
    madfifo->Write( (void *)tmp, ret*sizeof(unsigned char)); 
  } else {
    return 0;
  }
  //post("ReadMadUrl::wrote some data to fifo...");

  //memmove(mad_map, mad_map + STREAM_BUFFERSIZE - bytes_avail, bytes_avail);
  bytes_read = STREAM_BUFFERSIZE - bytes_avail;
  //post ("reading from MadUrl fifo...");
  madfifo->Read( (void *)mad_map,  bytes_read);
  //post ("read %d bytes from MadUrl fifo...", bytes_read);
  map_offset += (STREAM_BUFFERSIZE - bytes_avail);
  bytes_avail += bytes_read;

  //post("initial bytes_avail = %d", bytes_avail);
  offset = find_initial_frame( mad_map,
			       bytes_avail < STREAM_BUFFERSIZE ? bytes_avail :
			       STREAM_BUFFERSIZE);
 
  //crashed here sometimes !!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!

  //post("ReadMadUrl::found initial frame...offset %d", offset);

  pthread_create(&x_childthread, 0, fill_infifo, this);


  //offset =0;
  if ( offset < 0) {
    post("mad_open() couldn't find valid MPEG header\n");
    offset = 0;
    return false;
  }
  if ( offset > bytes_avail) {
    //seekable = 1;
    //post("Need to refill buffer (offset = %d)", offset);
    if ( !fill_buffer())
      { socket_disconnect(mad_fd); mad_fd = -1; return 0;}

    mad_stream_buffer(&stream, mad_map, bytes_avail);
  } else {
    mad_stream_buffer(&stream, mad_map + offset, bytes_avail - offset);
    bytes_avail -= offset;
  }	
  
  if ((mad_frame_decode(&frame, &stream) != 0)) {
    //alsaplayer_error("MAD error: %s", error_str(data->stream.error, data->str));
    switch (stream.error) {
    case MAD_ERROR_BUFLEN:
      post( "MAD_ERROR_BUFLEN");
      return false; //return 0;
    case MAD_ERROR_LOSTSYNC:
      if (mad_header_decode(&frame.header, &stream) == -1) {
	post("Invalid header (%s)", hostname);
      }
      mad_stream_buffer(&stream, stream.this_frame,
			bytes_avail - (stream.this_frame - mad_map));
      bytes_avail -= (stream.this_frame - mad_map);
      //post("avail = %d", data->bytes_avail);
      mad_frame_decode(&frame, &stream);
      break;
    case MAD_ERROR_BADBITALLOC:
      post( "MAD_ERROR_BADBITALLOC");
      return false;//return 0;
    case MAD_ERROR_BADCRC:
      post("MAD_ERROR_BADCRC"); //error_str( stream.error, str));
    case 0x232:				
    case 0x235:
      break;
    default:
      post("No valid frame found at start (pos: %d, error: 0x%x --> %x %x %x %x) (%s)", 
	   offset, 
	   stream.error, 
	   stream.this_frame[0],
	   stream.this_frame[1],
	   stream.this_frame[2],
	   stream.this_frame[3], fullurl);
      return false;//return 0;
    }
  }
  
  //if ( stream.error != MAD_ERROR_LOSTSYNC)
  
  num_channels = ( frame.header.mode == MAD_MODE_SINGLE_CHANNEL) ? 1 : 2;
  samplerate = (double)frame.header.samplerate;
  bitrate = frame.header.bitrate;
  mad_synth_frame (&synth, &frame);

  /* Calculate some values */
  bytes_avail = stream.bufend - stream.next_frame;
  seekable = 0; // Given really  
  mad_init = 1;
  //fill_buffer();
  return true;//return 1;
}


int ReadMadUrl::strip_shout_header(char *head, int n) {
  int i;
  for (i = 0; i < (n - 2); i++) {
    if (head[i] == 10 && head[i + 1] == 13)
      break;
    if (head[i] == '\n' && head[i + 1] == '\n')
      break;
  }
  head[i + 1] = '\0';
  return n - (i + 1);
}

int ReadMadUrl::strip_ice_header(char *head, int n) {
  int i;
  for (i = 0; i < (n - 2); i++) {
    if ((head[i] == '\n') && (head[i + 1] == '\n'))
      break;
  }
  head[i + 1] = '\0';
  return n - (i + 1);
}


void ReadMadUrl::socket_disconnect( int fd) {
  sys_closesocket(fd);
  post("connection closed");
}

/* connect to shoutcast server */
int ReadMadUrl::socket_connect(char *hostname, char *mountpoint, int portno) {
  struct          sockaddr_in server;
  struct          hostent *hp;

  /* variables used for communication with server */
  char            *sptr = NULL;
  char            request[STRBUF_SIZE];           /* string to be send to server */
  char            *url;           /* used for relocation */
  char*			bcname;         /* name of broadcast */
  char*			bcurl;          /* url of broadcast */
  char*			bcgenre;        /* genre of broadcast */
  char*			bcaim;          /* aim of broadcast */
  fd_set          fdset;
  struct timeval  tv;
  int           sockfd;                         /* socket to server */
  int           relocate;
  int           i, ret, rest;
  char            *cpoint = NULL;
  
  sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sockfd < 0) {
    error("ReadMadUrl: internal error while attempting to open socket");
    return (-1);
  }

  /* connect socket using hostname provided in command line */
  server.sin_family = AF_INET;
  hp = gethostbyname(hostname);
  if (hp == 0) {
    post("ReadMadUrl: bad host?");
    sys_closesocket(sockfd);
    return (-1);
  }
  memcpy((char *)&server.sin_addr, (char *)hp->h_addr, hp->h_length);

  /* assign client port number */
  server.sin_port = htons((unsigned short)portno);
  
  /* check mountpoint :: SHOUcast does not support mountpoints */
  if(strstr(mountpoint, "listen.pls") || strstr(mountpoint, "/")) {
    mountpoint = "";	/* the '/' is not part of the mountpoint !!! */
  }
  
  /* try to connect.  */
  post("connecting to http://%s:%d/%s", hostname, portno, mountpoint);
  if (connect(sockfd, (struct sockaddr *) &server, sizeof (server)) < 0) {
    error("connection failed!\n");
    sys_closesocket(sockfd);
    return (-1);
  }

  /* sheck if we can read/write from/to the socket */
  FD_ZERO( &fdset);
  FD_SET( sockfd, &fdset);
  tv.tv_sec  = 0;            /* seconds */
  tv.tv_usec = 500;        /* microseconds */
  
  ret = select(sockfd + 1, &fdset, NULL, NULL, &tv);
  if(ret != 0) {
    error("ReadMadUrl: can not read from socket");
    sys_closesocket(sockfd);
    return (-1);
  }

  /* build up stuff we need to send to server */
  sprintf(request, "GET /%s HTTP/1.0 \r\nHost: %s\r\nUser-Agent: ReadMadUrl 0.5\r\nAccept: */*\r\n\r\n", 
	  mountpoint, hostname);
  
  if(send(sockfd, request, strlen(request), 0) < 0) {
    post("ReadMadUrl: could not contact server... ");
#ifndef NT
    perror("send");
#endif
    return (-1);
  }

  relocate = false;
  
  if( recv(sockfd, request, STRBUF_SIZE, 0)<0) {
    error("ReadMadUrl: no response from server");
#ifndef NT
    perror( "recv" );
#endif
    socket_disconnect(sockfd);
    return (-1);
  }
  
  /* time to parse content of the response... */
  if(strstr(request, "audio/x-scpls") )    /* SHOUTcast playlist */
    {
      /* LATER: parse playlist and retry */
      post("SHOUTcast server returned a playlist, quitting");
      socket_disconnect(sockfd);
      return (-1);
    } 
  if(strstr(request, "HTTP"))    /* seems to be IceCast server */
    {
      strip_ice_header(request, STRBUF_SIZE);    
      if( (sptr = strstr(request, "302")) ) 
        {
	  cpoint = strstr(request, "Location:");
	  url = STRDUP(cpoint + 10);
	  post("relocation requested, try \"%s\" instead", url);
	  socket_disconnect(sockfd);
	  // return (shoutamp_connect_url(x, gensym(url)));
	  relocate = true;
	  return (-1);
        }
      if(!(sptr = strstr(request, "200")) && !relocate ) 
        {
	  error("ReadMadUrl: cannot connect to the (default) stream");
	  socket_disconnect(sockfd);
	  return (-1);
        }

      //post("ReadMadUrl: IceCast server detected");
      
      /* eventually read the rest of server's announcement */
      while((!(cpoint = strstr(request, "x-audiocast-public") ) ||
	     !(cpoint = strstr(request, "x-audiocast-mount") ) ||
	     !(cpoint = strstr(request, "x-audiocast-server-url") ) ||
	     !(cpoint = strstr(request, "x-audiocast-location") ) ||
	     !(cpoint = strstr(request, "x-audiocast-admin") ) ||
	     !(cpoint = strstr(request, "x-audiocast-name") ) ||
	     !(cpoint = strstr(request, "x-audiocast-genre") ) ||
	     !(cpoint = strstr(request, "x-audiocast-url") ) ||
	     !(cpoint = strstr(request, "x-audiocast-public") ) ||
	     !(cpoint = strstr(request, "x-audiocast-bitrate") ) ) &&
	    ( ret < STRBUF_SIZE ) )
        {
	  if( (rest = recv(sockfd, request+ret, STRBUF_SIZE-ret, 0) ) < 0 )
	    {
	      error("ReadMadUrl: no response from server");
#ifndef NT
	      perror("recv");
#endif
	      socket_disconnect(sockfd);
	      return(-1);
	    }
	  ret += rest;
        }
      
      // post("ReadMadUrl: server's header : %s", request );
      
      /* check what we got */
      if( (cpoint = strstr(request, "x-audiocast-mount:")))
        {
	  mountpoint = STRDUP(cpoint + 18);
	  for ( i=0; i<(int)strlen(mountpoint); i++ )
	    {
	      if(mountpoint[i] == '\n' )
                {
		  mountpoint[i] = '\0';
		  break;
                }    
	    }
	  //post("           mountpoint: %s", mountpoint);
        }
      if( (cpoint = strstr(request, "x-audiocast-server-url:")))
        {
	  sptr = STRDUP( cpoint + 24);
	  for ( i=0; i<(int)strlen(sptr); i++ )
	    {
	      if ( sptr[i] == '\n' )
                {
		  sptr[i] = '\0';
		  break;
                }    
	    }
	  //post("           server-url: %s", sptr);
	}
      if( (cpoint = strstr(request, "x-audiocast-location:")) )
	{
	  sptr = STRDUP( cpoint + 22);
	  for ( i=0; i<(int)strlen(sptr); i++ )
	    {
	      if ( sptr[i] == '\n' )
                {
		  sptr[i] = '\0';
		  break;
                }    
	    }
	  //post("           location: %s", sptr);
	}
      if( (cpoint = strstr(request, "x-audiocast-admin:")) )
	{
	  sptr = STRDUP( cpoint + 19);
	  for ( i=0; i<(int)strlen(sptr); i++ )
	    {
	      if ( sptr[i] == '\n' )
                {
		  sptr[i] = '\0';
		  break;
                }    
	    }
	  //post("           admin: %s", sptr);
	}
      if( (cpoint = strstr(request, "x-audiocast-name:")) )
	{
	  bcname = STRDUP( cpoint + 17);
	  for ( i=0; i<(int)strlen(bcname); i++ )
	    {
	      if ( bcname[i] == '\n' )
                {
		  bcname[i] = '\0';
		  break;
                }    
	    }
	  //post("           name: %s", bcname);
	}
      if( (cpoint = strstr(request, "x-audiocast-genre:")) )
	{
	  bcgenre = STRDUP( cpoint + 18);
	  for ( i=0; i<(int)strlen(bcgenre); i++ )
	    {
	      if ( bcgenre[i] == '\n' )
                {
		  bcgenre[i] = '\0';
		  break;
                }    
	    }
	  //post("           genre: %s", bcgenre);
	}
      if( (cpoint = strstr(request, "x-audiocast-url:")) )
	{
	  bcurl = STRDUP( cpoint + 16);
	  for ( i=0; i<(int)strlen(bcurl); i++ )
	    {
	      if ( bcurl[i] == '\n' )
                {
		  bcurl[i] = '\0';
		  break;
                }    
	    }
	  //post("           url: %s", bcurl);
	}
      if( (cpoint = strstr(request, "x-audiocast-public:1")) )
	{
	  //post("           broadcast is public");
	}
      else if( (cpoint = strstr(request, "x-audiocast-public:0")) )
	{
	  //post("           broadcast is NOT public");
	}
      if( (cpoint = strstr(request, "x-audiocast-bitrate:")) )
	{
	  sptr = STRDUP( cpoint + 20);
	  for ( i=0; i<(int)strlen(sptr); i++ )
	    {
	      if ( sptr[i] == '\n' )
                {
		  sptr[i] = '\0';
		  break;
                }    
	    }
	  post("bitrate: %s", sptr);
	}
      if( (cpoint = strstr(request, "x-audiocast-udpport:")) )
	{
	  post("sorry, server wants UDP connection!");
	  return (-1);
	} 
    }
  else    /* it is a SHOUTcast server */
    {
      strip_shout_header (request, STRBUF_SIZE);
      //post("ReadMadUrl: SHOUTcast server detected");
      if(strncmp(request, "ICY 401", 7) == 0)
        {
	  post("ICY 401 Service Unavailable");
	  return (-1);
        }
      if(strncmp(request, "ICY 200 OK", 10) == 0)
        {
	  /* recv and decode info about broadcast line by line */
	  post(" connecting to stream...");
	  i = ret;
	  /* break response down to single lines */
	  while(i < STRBUF_SIZE - 1)
	    {
	      for(rest = ret; rest < STRBUF_SIZE - 1; rest++)
                {
		  if(recv(sockfd, request+rest, 1, 0)!=1)
		    break;
		  if (request[rest] == '\n')
		    break;
		  if (request[rest] == '\r')
		    break;
		  i++;
                }
	      request[rest] = '\0';
	      
	      /* check what we got */
	      if(!strncmp(request, "icy-name:", 9))
                {
		  bcname = STRDUP(request + 9);
		  //post("           name: %s", bcname);
                }
	      else if(!strncmp(request, "x-audiocast-name:", 17))
                {
		  bcname = STRDUP(request + 17);
		  //post("           name: %s", bcname);
                }
	      if(!strncmp(request, "icy-genre:", 9))
                {
		  bcgenre = STRDUP(request + 10);
		  //post("           genre: %s", bcgenre);
                }
	      if(!strncmp(request, "icy-aim:", 7))
                {
		  bcaim = STRDUP(request + 8);
		  //post("           aim: %s", bcaim);
                }
	      if(!strncmp(request, "icy-url:", 7))
                {
		  bcurl = STRDUP(request + 8);
		  //post("           url: %s", bcurl);  
                }
	      if(!strncmp(request, "icy-pub:1", 8))
                {
		  //post("           broadcast is public");
                }
	      else if(!strncmp(request, "icy-pub:0", 8))
                {
		  //post("           broadcast is NOT public");
                }
	      if(!strncmp(request, "icy-br:", 6))
                {
		  sptr = STRDUP(request + 7);
		  post("bitrate: %s", sptr);
                }
	      if(!strncmp(request, "x-audiocast-udpport:", 20))
                {
		  post("sorry, server wants UDP connection!");
		  return (-1);
                } 
	      
            }
        }
      else 
        {
	  post("unknown response from server");
	  return (-1);
        }
      relocate = false;    
    }
  if(relocate)
    {
      error("ReadMadUrl: too many HTTP relocations");
      return (-1);
    }
  post("connected to http://%s:%d/%s", hp->h_name, portno, mountpoint);
  
  return (sockfd);
}



/* connect using url like "http://localhost:8000/mountpoint" */
void ReadMadUrl::set_url( const char *url) {
  char *hname, *tmpport;
  char *h, *p;
  char *hostptr;
  char *r_hostptr;
  char *pathptr;
  char *portptr;
  char *p0;
  char *defaultportstr = "8000";
  int stringlength;
  int portno;
  
  /* strip http:// or ftp:// */
  p = (char *)url;
  if (strncmp(p, "http://", 7) == 0)
    p += 7;
  
  if (strncmp(p, "ftp://", 6) == 0)
    p += 6;
  
  hostptr = p;
  while (*p && *p != '/')	/* look for end of hostname:port */
    p++;
  p++;					/* also skip '/' */
  pathptr = p;
  
  r_hostptr = --p;
  while (*p && hostptr < p && *p != ':' && *p != ']')	/* split at ':' */
    p--;
  
  if (!*p || p < hostptr || *p != ':') {
    portptr = NULL;
  }
  else{
    portptr = p + 1;
    r_hostptr = p - 1;
  }
  if (*hostptr == '[' && *r_hostptr == ']') {
    hostptr++;
    r_hostptr--;
  }
  
  stringlength = r_hostptr - hostptr + 1;
  h = (char *)getbytes( stringlength + 1 );
  if (h == NULL) {
    hname = NULL;
    tmpport = NULL;
    pathptr = NULL; 
  }
  strncpy(h, hostptr, stringlength);
  *(h+stringlength) = '\0';
  hname = h;	/* the hostname */
  
  if (portptr) {
    stringlength = (pathptr - portptr);
    if(!stringlength) portptr = NULL;
  }
  if (portptr == NULL) {
    portptr = defaultportstr;
    stringlength = strlen(defaultportstr);
  }
  p0 = (char *)getbytes(stringlength + 1);
  if (p0 == NULL) {
    freebytes(h, stringlength + 1);
    hname = NULL;
    tmpport = NULL;
    pathptr = NULL;
  }
  strncpy(p0, portptr, stringlength);
  *(p0 + stringlength) = '\0';
  
  for (p = p0; *p && isdigit((unsigned char) *p); p++) ;
  
  *p = '\0';
  tmpport = (char *) p0;
  /* convert port from string to int */
  portno = (int)strtol(tmpport, NULL, 10);
  freebytes(p0, stringlength + 1);
  
  
  strcpy(hostname, hname);
  strcpy(mountpoint, pathptr);
  //hostname = hname;
  //mountpoint = pathptr;
  port = portno;
  
}

#endif
