/*
 * readanysf~  external for pd. 
 * 
 * Copyright (C) 2003 August Black
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ReadVorbisUrl.cpp || much code (C) by Olaf Matthes
 * http://www.akustische-kunst.org/puredata/shout/shout.html
 */

#ifdef READ_VORBIS_URL

#include "ReadVorbisUrl.h"
#include <m_pd.h>



void  *ReadVorbisUrl::fill_infifo(void *zz ) {

  int ret, wret, tmpfd, count=0;
  unsigned char tmp[VSOCKET_READSIZE];
  ReadVorbisUrl *rvu = (ReadVorbisUrl *)zz;

  while (1){ 
  
    pthread_mutex_lock( rvu->get_mutex() );
    if ( rvu->get_quit() ) {
      pthread_mutex_unlock( rvu->get_mutex() );
      break;
    }

    while ( rvu->get_fifo()->FreeSpace() > VSOCKET_READSIZE+10) {  
      if ( rvu->get_quit() ) {
	break;
      }
      tmpfd = rvu->get_fd();
      rvu->set_cachesize( ); //(float)rvu->get_fifo()->madfifo->UsedSpace();  
      pthread_mutex_unlock( rvu->get_mutex() );
      ret =-1;
      ret = recv( tmpfd , tmp, VSOCKET_READSIZE, 0 );
      if(ret < 0) {
	post("receive error" );
      }

      //post("receiving data");
      pthread_mutex_lock( rvu->get_mutex() );

      if (ret > 0) {
	wret = rvu->get_fifo()->Write( (void *)tmp, ret*sizeof(unsigned char)); 
      } else {
	bzero((void *)tmp, sizeof(tmp)); //write silence to the fifo
	wret = rvu->get_fifo()->Write( (void *)tmp, ret*sizeof(unsigned char)); 
	if (count++ > 5) {
	  count = 0;
	  sys_closesocket(tmpfd);
	  post( "closed the connection, reconnecting to %s, %s, %d", 
		rvu->get_hostname(), 
		rvu->get_mountpoint(), 
		rvu->get_port() );
	  if (( tmpfd = rvu->socket_connect( rvu->get_hostname(), 
					     rvu->get_mountpoint(), 
					     rvu->get_port() ) ) < 0) {	  
	    post("couldn't reconnect,...trying again");
	  }
	  rvu->set_fd(tmpfd);
	} else {
	  post("couldn't get data, waiting %d seconds.", 6-count);
	  break; //or sleep
	}
      }
    }
    if ( rvu->get_quit() ) {
      //post("got quit,...now quitting.");
      pthread_mutex_unlock( rvu->get_mutex() );
      break;
    }
    pthread_mutex_unlock( rvu->get_mutex() );
    //post("fifo full in thread.. waiting");
    sleep(1);//pthread_cond_wait( rvu->get_condition() , rvu->get_mutex() );
  }


  post("exiting thread");
  pthread_cond_signal(rvu->get_condition());
  return NULL;
}

ReadVorbisUrl::ReadVorbisUrl() {
  vorbis_fd = -1;
  port = 0;
  quit = false;
  init = false;

  vfifo = new Fifo( VSTREAM_FIFOSIZE );

  callbacks.read_func = read_func;
  callbacks.seek_func = seek_func;
  callbacks.close_func = close_func;
  callbacks.tell_func = tell_func;

  pthread_mutex_init(&x_mutex, 0);
  pthread_cond_init(&x_condition, 0);

}

ReadVorbisUrl::~ReadVorbisUrl(){

  if (init){
    pthread_mutex_lock(&x_mutex);
    quit = true;
    pthread_mutex_unlock(&x_mutex);
    
    
    post("signalling...");
    pthread_cond_signal(&x_condition);
    
    pthread_cond_wait(&x_condition  , &x_mutex );
    //post("after waiting...");


    if ( ov_clear(&vf) != 0) 
      post("couldn't deconstruct the vorbis file reader");
    //if (fp != NULL) { //don't know why, but this will cause a segfault
    // fclose(fp); 
    //}
  }
  pthread_cond_destroy(&x_condition);
  pthread_mutex_destroy(&x_mutex);

}


bool ReadVorbisUrl::Initialize(const char *fullurl) {  

  vorbis_info *vi;
  unsigned char inbuffer[8500*2];

  if (vorbis_fd != -1) {
    close(vorbis_fd); 
  }

  set_url(fullurl);
    
  if (( vorbis_fd = socket_connect( hostname, mountpoint, port)) < 0) {
    post( "Initialized() failed");
    return false;  //return 0;
  }
  //post("RVU connected: vorbis_fd %d", vorbis_fd);

  init = true;

  int n =0;
  while ( n < 8500 ) {
    int ret = recv( vorbis_fd, inbuffer, 1024, 0);
    if(ret < 0) 
      post("receive error:stream %d,  ret %d", vorbis_fd, ret);
       
    vfifo->Write((void *)inbuffer, ret*sizeof(unsigned char));
    //post("wrote %d chars to vfifo, %d", wret, vfifo->UsedSpace() );
    n += ret;
  }

  pthread_create(&x_readthread, 0, fill_infifo, this);

  if( ov_open_callbacks( (void *)vfifo, &vf, NULL, 0, callbacks) < 0 ) {
    fprintf(stderr," does not appear to be an Ogg bitstream.\n");
    return false;
  }

  //post("RVU open_callbacks done...");

  vi=ov_info(&vf,-1);
  samplerate = (double)vi->rate;
  num_channels = vi->channels;
  lengthinseconds = 0.0;//(double)ov_time_total(&vf, -1);
  //post("ReadVorbisUrl: opening url: [%s] %ld (Hz), %d chan(s)", 
  //   fullurl, vi->rate, vi->channels);


  return true;
}

int ReadVorbisUrl::Decode(float *buffer, int size) {
  long ret = 0;
  int current_section;
  float **buftmp;
  int x=0;
  if (CHUNKSIZE > (unsigned int)size) return 0;

  pthread_mutex_lock(&x_mutex );
  ret = ov_read_float(&vf, &buftmp , CHUNKSIZE,  &current_section);
  pthread_mutex_unlock( &x_mutex );


  if (ret <= 0) {
    //post("ov_read_float is zero");
    for(int j = 0; j < 1024; j++)
      {
	buffer[x++] = 0.0;
      }
  } else {

    for(int j = 0; j < ret; j++) {
      buffer[x++] = buftmp[0][j];
      if (num_channels == 2) 
      	buffer[x++] = buftmp[1][j];
    }
  }
  //post("x %d", x);
  return x;  //ret;

}

size_t ReadVorbisUrl::read_func(void *ptr, size_t size, size_t nmemb, void *datasource) {

  Fifo * tmpfifo = ( Fifo *)datasource;
  unsigned int   ret = size*nmemb;
  
  //post("reading from fifo,  ret %d, size %d, nmemb %d, fifospace %d", 
  //ret,size,nmemb, tmpfifo->UsedSpace() );

  tmpfifo->Read( (void *)(ptr),  ret);

  //post("read from fifo,  ret %d, size %d, nmemb %d, fifospace %d", 
  //    ret,size,nmemb, tmpfifo->UsedSpace() );
  //if(ret <= 0) {
  // post("read error: ret %d", ret);
  //} 

  return ret;//size*nmemb;
}

int ReadVorbisUrl::seek_func(void *datasource, ogg_int64_t offset, int whence){
  return -1;
}

int ReadVorbisUrl::close_func(void *datasource) {
  int stream  = (int)datasource;
  return close(stream);
}

long ReadVorbisUrl::tell_func(void *datasource) {
  return -1;
}

bool ReadVorbisUrl::Rewind() {
    return false;
}

bool ReadVorbisUrl::PCM_seek(long bytes) {
    return false;
}

bool ReadVorbisUrl::TIME_seek(double seconds) {
    return false;
}

int ReadVorbisUrl::check_for_data(int sock) {
  fd_set set;
  struct timeval tv;
  t_int ret;
  
  tv.tv_sec = 0;
  tv.tv_usec = 20000;
  FD_ZERO(&set);
  FD_SET(sock, &set);
  ret = select(sock + 1, &set, NULL, NULL, &tv);
  if (ret > 0)
    return 1;
  return 0;
}

int ReadVorbisUrl::socket_connect(char *hostname, char *mountpoint, int portno) {
  struct  sockaddr_in server;
  struct  hostent *hp;
  
  /* variables used for communication with server */
  char            *sptr = NULL;
  char            request[STRBUF_SIZE];           /* string to be send to server */
  //char            *url;               /* used for relocation */
  fd_set          fdset;
  struct timeval  tv;
  t_int           sockfd;                         /* socket to server */
  //t_int           relocate; 
  //t_int numrelocs = 0;
  t_int           i, ret; //rest, nanswers=0;
  //char            *cpoint = NULL;
  t_int           eof = 0;
  
  sockfd = socket(AF_INET, SOCK_STREAM, IPPROTO_TCP);
  if (sockfd < 0) {
    error(" : internal error while attempting to open socket");
    return (-1);
  }
  
  /* connect socket using hostname provided in command line */
  server.sin_family = AF_INET;
  hp = gethostbyname(hostname);
  if (hp == 0)    {
    post(" : bad host?");
    sys_closesocket(sockfd);
    return (-1);
  }
  memcpy((char *)&server.sin_addr, (char *)hp->h_addr, hp->h_length);
  
  /* assign client port number */
  server.sin_port = htons((unsigned short)portno);
  
  /* try to connect.  */
  //post(" : connecting to http://%s:%d/%s", hostname, portno, mountpoint);
  if (connect(sockfd, (struct sockaddr *) &server, sizeof (server)) < 0)    {
    error(" : connection failed!\n");
    sys_closesocket(sockfd);
    return (-1);
  }
  
  /* sheck if we can read/write from/to the socket */
  FD_ZERO( &fdset);
  FD_SET( sockfd, &fdset);
  tv.tv_sec  = 0;            /* seconds */
  tv.tv_usec = 500;        /* microseconds */
  
  ret = select(sockfd + 1, &fdset, NULL, NULL, &tv);
  if(ret != 0) {
    error(" : can not read from socket");
    sys_closesocket(sockfd);
    return (-1);
  }
  
  /* build up stuff we need to send to server */
  sprintf(request, "GET /%s HTTP/1.0 \r\nHost: %s\r\nUser-Agent:   0.2\r\nAccept: audio/x-ogg\r\n\r\n", 
	  mountpoint, hostname);
  
  if ( send(sockfd, request, strlen(request), 0) < 0 )    /* say hello to server */
    {
      post(" : could not contact server...");
      return (-1);
    }
  
		/* read first line of response */
  i = 0;
  while(i < STRBUF_SIZE - 1) {
    if( check_for_data(sockfd) ) {
      if (recv(sockfd, request + i, 1, 0) <= 0) {
	error(" : could not read from socket, quitting");
	sys_closesocket(sockfd);
	return (-1);
      }
      if (request[i] == '\n')
	break;
      if (request[i] != '\r')
	i++;
    }
  }
  request[i] = '\0';
  
  /* time to parse content of the response... */
  if(strstr(request, "HTTP/1.0 200 OK"))    /* server is ready */
    {
      //post(" : IceCast2 server detected");
      while(!eof)		/* read everything we can get */
	{
	  i = 0;
	  while(i < STRBUF_SIZE - 1)
	    {
	      if( check_for_data(sockfd) )
		{
		  if (recv(sockfd, request + i, 1, 0) <= 0)
		    {
		      error(" : could not read from socket, quitting");
		      sys_closesocket(sockfd);
		      return (-1);
		    }
		  if(request[i] == '\n')  /* leave at end of line */
		    break;
		  if(request[i] == 0x0A)  /* leave at end of line */
		    break;
		  if(request[i] != '\r')	/* go on until 'return' */
		    i++;
		}
	    }
	  request[i] = '\0';	/* make it a null terminated string */
	  
	  if( ( sptr = strstr(request, "application/x-ogg") ) ) 
	    {		/* check for content type */
	      //post(" : Ogg Vorbis stream found");
	    }
	  if( ( sptr = strstr(request, "ice-name:") ) ) 
	    {		/* display ice-name */
	      //post(" : \"%s\"", sptr + 10);
	    }
	  if(i == 0)eof = 1;	/* we got last '\r\n' from server */
	}
    }
  else    /* wrong server or wrong answer */
    {
      post(" : unknown response from server");
      sys_closesocket(sockfd);
      return (-1);
    }
  //post(" : connected to http://%s:%d/%s", hp->h_name, portno, mountpoint);
  
  return (sockfd);
}


void ReadVorbisUrl::socket_disconnect( int fd) {
  sys_closesocket(fd);
  post("connection closed");
}


void ReadVorbisUrl::set_url(const char *url) {
  char *hname, *tmpport;
  char *h, *p;
  char *hostptr;
  char *r_hostptr;
  char *pathptr;
  char *portptr;
  char *p0;
  char *defaultportstr = "8000";
  t_int stringlength;
  t_int portno;

  /* strip http:// or ftp:// */
  p = (char *)url;
  if (strncmp(p, "http://", 7) == 0)
    p += 7;

  if (strncmp(p, "ftp://", 6) == 0)
    p += 6;
  
  hostptr = p;
  while (*p && *p != '/')	/* look for end of hostname:port */
    p++;
  p++;					/* also skip '/' */
  pathptr = p;
  
  r_hostptr = --p;
  while (*p && hostptr < p && *p != ':' && *p != ']')	/* split at ':' */
    p--;
  
  if (!*p || p < hostptr || *p != ':') {
    portptr = NULL;
  }
  else{
    portptr = p + 1;
    r_hostptr = p - 1;
  }
  if (*hostptr == '[' && *r_hostptr == ']') {
    hostptr++;
    r_hostptr--;
  }
  
  stringlength = r_hostptr - hostptr + 1;
  h = (char *)getbytes( stringlength + 1);
  if (h == NULL) {
    hname = NULL;
    tmpport = NULL;
    pathptr = NULL; 
  }
  strncpy(h, hostptr, stringlength);
  *(h+stringlength) = '\0';
  hname = h;	/* the hostname */
  
  if (portptr) {
    stringlength = (pathptr - portptr);
    if(!stringlength) portptr = NULL;
  }
  if (portptr == NULL) {
    portptr = defaultportstr;
    stringlength = strlen(defaultportstr);
  }
  p0 = (char *)getbytes(stringlength + 1);
  if (p0 == NULL) {
    freebytes(h, stringlength + 1);
    hname = NULL;
    tmpport = NULL;
    pathptr = NULL;
  }
  strncpy(p0, portptr, stringlength);
  *(p0 + stringlength) = '\0';
  
  for (p = p0; *p && isdigit((unsigned char) *p); p++) ;
  
  *p = '\0';
  tmpport = (char *) p0;
  /* convert port from string to int */
  portno = (int)strtol(tmpport, NULL, 10);
  freebytes(p0, stringlength + 1);

  strcpy(hostname, hname);
  strcpy(mountpoint, pathptr);
  //hostname = hname;
  //mountpoint = pathptr;
  port = portno;
}


#endif
