/*
 * readanysf~  external for pd. 
 * 
 * Copyright (C) 2003 August Black
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * ReadVorbisUrl.h
 */

#ifdef READ_VORBIS_URL

#ifndef _READVORBISURL_H_
#define _READVORBISURL_H_

#include "Readsf.h"
#include "Fifo.h"
#include "generic.h"

extern "C" {
#include <vorbis/codec.h>
#include <vorbis/vorbisfile.h>
}

#include <sys/types.h>
#include <ctype.h>
#include <string.h>
#include <stdio.h>
#include <errno.h>
#include <fcntl.h>
#include <pthread.h>

//#include <stdlib.h>
//#include <math.h>

#ifdef NT
#include <io.h> /* for 'write' in pute-function only */
#include <winsock.h>
#include <winbase.h>
#else
#include <sys/socket.h>
#include <netinet/in.h>
#include <netinet/tcp.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <unistd.h>
#define SOCKET_ERROR -1
#endif

#ifdef NT
#pragma warning( disable : 4244 )
#pragma warning( disable : 4305 )
#endif



#ifdef NT
#define     sys_closesocket closesocket
#else
#define     sys_closesocket close
#endif

#define VSTREAM_FIFOSIZE (32 * 1024)
#define VSTREAM_BUFFERSIZE (4 * 1024)
#define VSOCKET_READSIZE 1024
#define     STRBUF_SIZE             1024


class ReadVorbisUrl : public Readsf {
 private:
  pthread_mutex_t x_mutex;
  pthread_cond_t x_condition;
  pthread_t x_readthread;

  bool quit,init;

  OggVorbis_File vf;
  ov_callbacks callbacks;
  int vorbis_fd;
  Fifo *vfifo;

  char hostname[1024];
  char mountpoint[1024];
  int port;
  
  void socket_disconnect(int fd);
  void set_url(const char *url);
  int socket_connect(char *hostname, char *mountpoint, int portno);
  int check_for_data(int sock);

  static void *fill_infifo(void *leer);

  static size_t read_func(void *ptr, size_t size, size_t nmemb, void *datasource);
  static int seek_func(void *datasource, ogg_int64_t offset, int whence);
  static int close_func(void *datasource);
  static long tell_func(void *datasource);


 public:
  ReadVorbisUrl();
  virtual ~ReadVorbisUrl();
  virtual bool Initialize(const char *filename);
  virtual int Decode(float *buffer, int size);
  virtual bool Rewind();
  virtual bool PCM_seek(long bytes);
  virtual bool TIME_seek(double seconds);

  pthread_mutex_t *get_mutex() { return &x_mutex;}
  pthread_cond_t *get_condition(){ return &x_condition;}
  int get_fd() { return vorbis_fd;}
  void set_fd(int tmpfd) { vorbis_fd = tmpfd;}
  void set_cachesize() { cachesize = vfifo->UsedSpace();}


  char * get_hostname() { return hostname;}
  char * get_mountpoint() { return mountpoint;}
  int get_port() { return port;}
  bool get_quit() { return quit;}
  Fifo * get_fifo() { return vfifo;}

};

#endif
#endif //ifdef READVORBIS
